/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

/* This file implements the nsIMsgCloudFileProvider interface.
 *
 * This component handles the Dropbox implementation of the
 * nsIMsgCloudFileProvider interface.
 */

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cu = Components.utils;
const Cr = Components.results;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource:///modules/OAuth2.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource:///modules/gloda/log4moz.js");
Cu.import("resource:///modules/cloudFileAccounts.js");
Cu.import("resource://gre/modules/Http.jsm");

const kBadAccessToken = 401;
const kAuthSecretRealm = "Dropbox Auth Secret";
// According to Dropbox, the kMaxFileSize is a fixed limit.
const kMaxFileSize = 157286400;
const kUserInfoPath = "users/get_space_usage";
const kDeletePath = "fileops/delete/?root=sandbox";
const kSharesPath = "sharing/create_shared_link_with_settings";
const kListSharesPath = "sharing/list_shared_links";
const kFilesPutPath = "files_put/sandbox/";

const kAuthBaseUrl = "https://www.dropbox.com/oauth2/";
const kScope = null;
const kAuthUrl = "authorize";
const kTokenUrl = "token";

var gServerUrl = "https://api.dropbox.com/2/";
var gContentUrl = "https://api-content.dropbox.com/1/";
var gAuthUrl = "https://www.dropbox.com/1/";
var gLogoutUrl = "https://www.dropbox.com/logout";
const kClientId = "7topeaxs06u740q";
const kClientSecret = "jto5staa22t83ea";

function wwwFormUrlEncode(aStr) {
  return encodeURIComponent(aStr).replace(/!/g, '%21')
                                 .replace(/'/g, '%27')
                                 .replace(/\(/g, '%28')
                                 .replace(/\)/g, '%29')
                                 .replace(/\*/g, '%2A');
}


function nsDropbox() {
  this.log = Log4Moz.getConfiguredLogger("Dropbox");
  this.log.level = Log4Moz.Level.All;
  this.log.addAppender(new Log4Moz.ConsoleAppender(new Log4Moz.BasicFormatter()));
  this._oauth = new OAuth2(kAuthBaseUrl, kScope, kClientId, kClientSecret);
  this._oauth.authURI = kAuthBaseUrl + kAuthUrl;
  this._oauth.tokenURI = kAuthBaseUrl + kTokenUrl;

  let account = this;
  Object.defineProperty(this._oauth, "accessToken", {
    get: function getAccessToken() {
      if (!this.mAccessToken) {
        let authToken = cloudFileAccounts.getSecretValue(account.accountKey,
                                                         cloudFileAccounts.kTokenRealm);
        this.mAccessToken = authToken || "";
      }
      return this.mAccessToken;
    },
    set: function setAccessToken(aVal) {
      if (!aVal)
        aVal = "";

      cloudFileAccounts.setSecretValue(account.accountKey,
                                       cloudFileAccounts.kTokenRealm,
                                       aVal);

      return (this.mAccessToken = aVal);
    },
    enumerable: true
  });
}

nsDropbox.prototype = {
  /* nsISupports */
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIMsgCloudFileProvider, Ci.nsIMsgCloudFileProvider2]),

  classID: Components.ID("{2fd8a64a-a496-4cf4-9d6b-d3f9800c6322}"),

  get type() "Dropbox",
  get displayName() "Dropbox",
  get serviceURL() "https://www.dropbox.com/",
  get iconClass() "chrome://pbfilelink/content/dropbox/dropbox.png",
  get accountKey() this._accountKey,
  get lastError() this._lastErrorText,
  get settingsURL() "chrome://pbfilelink/content/dropbox/settings.xhtml",
  get managementURL() "chrome://pbfilelink/content/dropbox/management.xhtml",

  _accountKey: false,
  _prefBranch: null,
  // If an access token exists, the user is logged in.
  get _loggedIn() !!this._oauth.accessToken,
  _authToken: "",
  _userInfo: null,
  _file : null,
  _requestDate: null,
  _successCallback: null,
  _connection: null,
  _request: null,
  _uploadingFile : null,
  _uploader : null,
  _lastErrorStatus : 0,
  _lastErrorText : "",
  _maxFileSize : kMaxFileSize,
  _totalStorage: -1,
  _fileSpaceUsed : -1,
  _uploads: [],
  _urlsForFiles : {},
  _uploadInfo : {}, // upload info keyed on aFiles.

  /**
   * Initialize this instance of nsDropbox, setting the accountKey.
   *
   * @param aAccountKey the account key to initialize this provider with
   */
  init: function nsDropbox_init(aAccountKey) {
    this._accountKey = aAccountKey;
    this._prefBranch = Services.prefs.getBranch("mail.cloud_files.accounts." + 
                                                aAccountKey + ".");
  },

  /**
   * The callback passed to an nsDropboxFileUploader, which is fired when
   * nsDropboxFileUploader exits.
   *
   * @param aRequestObserver the request observer originally passed to
   *                         uploadFile for the file associated with the
   *                         nsDropboxFileUploader
   * @param aStatus the result of the upload
   */
  _uploaderCallback : function nsDropbox__uploaderCallback(aRequestObserver,
                                                           aStatus) {
    aRequestObserver.onStopRequest(null, null, aStatus);
    this._uploadingFile = null;
    this._uploads.shift();
    if (this._uploads.length > 0) {
      let nextUpload = this._uploads[0];
      this.log.info("chaining upload, file = " + nextUpload.file.leafName);
      this._uploadingFile = nextUpload.file;
      this._uploader = nextUpload;
      try {
        this.uploadFile(nextUpload.file, nextUpload.callback);
      }
      catch (ex) {
        nextUpload.callback(nextUpload.requestObserver, Cr.NS_ERROR_FAILURE);
      }
    }
    else
      this._uploader = null;
  },

  /** 
   * Attempts to upload a file to Dropbox.
   *
   * @param aFile the nsILocalFile to be uploaded
   * @param aCallback an nsIRequestObserver for listening for the starting
   *                  and ending states of the upload.
   */
  uploadFile: function nsDropbox_uploadFile(aFile, aCallback) {
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;

    this.log.info("uploading " + aFile.leafName);

    // Some ugliness here - we stash requestObserver here, because we might
    // use it again in _getUserInfo.
    this.requestObserver = aCallback;

    // if we're uploading a file, queue this request.
    if (this._uploadingFile && this._uploadingFile != aFile) {
      let uploader = new nsDropboxFileUploader(this, aFile,
                                               this._uploaderCallback
                                                   .bind(this),
                                               aCallback);
      this._uploads.push(uploader);
      return;
    }
    this._file = aFile;
    this._uploadingFile = aFile;

    let successCallback = this._finishUpload.bind(this, aFile, aCallback);
    if (!this._loggedIn)
      return this._logonAndGetUserInfo(successCallback, null, true);
    this.log.info("getting user info");
    if (!this._userInfo)
      return this._getUserInfo(successCallback);
    successCallback();
  },

  /**
   * A private function used to ensure that we can actually upload the file
   * (we haven't exceeded file size or quota limitations), and then attempts
   * to kick-off the upload.
   *
   * @param aFile the nsILocalFile to upload
   * @param aCallback an nsIRequestObserver for monitoring the starting and
   *                  ending states of the upload.
   */
  _finishUpload: function nsDropbox__finishUpload(aFile, aCallback) {
    let exceedsFileLimit = Ci.nsIMsgCloudFileProvider.uploadExceedsFileLimit;
    let exceedsQuota = Ci.nsIMsgCloudFileProvider.uploadWouldExceedQuota;
    if (aFile.fileSize > this._maxFileSize)
      return aCallback.onStopRequest(null, null, exceedsFileLimit);
    if (aFile.fileSize > this.remainingFileSpace)
      return aCallback.onStopRequest(null, null, exceedsQuota);

    delete this._userInfo; // force us to update userInfo on every upload.

    if (!this._uploader) {
      this._uploader = new nsDropboxFileUploader(this, aFile,
                                                 this._uploaderCallback
                                                     .bind(this),
                                                 aCallback);
      this._uploads.unshift(this._uploader);
    }

    this._uploadingFile = aFile;
    this._uploader.uploadFile();
  },

  /**
   * Attempts to cancel a file upload.
   *
   * @param aFile the nsILocalFile to cancel the upload for.
   */
  cancelFileUpload: function nsDropbox_cancelFileUpload(aFile) {
    if (this._uploadingFile.equals(aFile)) {
      this._uploader.cancel();
    }
    else {
      for (let i = 0; i < this._uploads.length; i++)
        if (this._uploads[i].file.equals(aFile)) {
          this._uploads[i].requestObserver.onStopRequest(
            null, null, Ci.nsIMsgCloudFileProvider.uploadCanceled);
          this._uploads.splice(i, 1);
          return;
        }
    }
  },

  /**
   * A private function used to retrieve the profile information for the
   * user account associated with the accountKey.
   *
   * @param successCallback the function called if information retrieval
   *                        is successful
   * @param failureCallback the function called if information retrieval fails
   */
  _getUserInfo: function nsDropbox__getUserInfo(successCallback, failureCallback) {
    let requestUrl = gServerUrl + kUserInfoPath;
    this.log.info("get_account_info requestUrl = " + requestUrl);

    if (!successCallback)
      successCallback = function() {
        this.requestObserver
            .onStopRequest(null, null,
                           this._loggedIn ? Cr.NS_OK : Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    if (!failureCallback)
      failureCallback = function () {
        this.requestObserver
            .onStopRequest(null, null, Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    let accountInfoSuccess = function(aResponseText, aRequest) {
      this.log.info("get_account_info request response = " + aResponseText);

      try {
        this._userInfo = JSON.parse(aResponseText);

        if (!this._userInfo) {
          this.failureCallback();
          return;
        }

        this._totalStorage = this._userInfo.allocation.allocated;
        this._fileSpaceUsed = this._userInfo.used;
        this.log.info("storage total = " + this._totalStorage);
        this.log.info("storage used = " + this._fileSpaceUsed);
        successCallback();
      }
      catch(e) {
        // most likely bad JSON
        this.log.error("Failed to parse account info response: " + e);
        this.log.error("Account info response: " + aResponseText);
        failureCallback();
      }
    }.bind(this);
    let accountInfoFailure = function(aException, aResponseText, aRequest) {
      if (aRequest.status == kBadAccessToken) {
        // what to do here?
        //this.log.info("got bad token");
        //this._loggedIn = false;
        //this._cachedAuthToken = "";
        //this._cachedAuthSecret = "";
        //successCallback();
        //return;
      }
      this.log.error("user info failed, status = " + aRequest.status);
      this.log.error("response text = " + aResponseText);
      this.log.error("exception = " + aException);
      failureCallback();
    }.bind(this)

    // Request to get user info
    httpRequest(requestUrl, {
                  onLoad: accountInfoSuccess,
                  onError: accountInfoFailure,
                  method: "POST",
                  headers: [["Authorization", "Bearer " + this._oauth.accessToken]]
                });
  },

  /**
   * A private function that first ensures that the user is logged in, and then
   * retrieves the user's profile information.
   *
   * @param aSuccessCallback the function called on successful information
   *                         retrieval
   * @param aFailureCallback the function called on failed information retrieval
   * @param aWithUI a boolean for whether or not we should display authorization
   *                UI if we don't have a valid token anymore, or just fail out.
   */
  _logonAndGetUserInfo: function nsDropbox_logonAndGetUserInfo(aSuccessCallback,
                                                               aFailureCallback,
                                                               aWithUI) {
    if (!aFailureCallback)
      aFailureCallback = function () {
        this.requestObserver
            .onStopRequest(null, null, Ci.nsIMsgCloudFileProvider.authErr);
      }.bind(this);

    return this.logon(function() {
      this._getUserInfo(aSuccessCallback, aFailureCallback);
    }.bind(this), aFailureCallback, aWithUI);
  },

  /**
   * For some nsILocalFile, return the associated sharing URL.
   *
   * @param aFile the nsILocalFile to retrieve the URL for
   */
  urlForFile: function nsDropbox_urlForFile(aFile) {
    return this._urlsForFiles[aFile.path];
  },

  /**
   * For some nsILocalFile, return the associated sharing URL.
   *
   * @param aFile the nsILocalFile to retrieve the URL for
   */
  urlForFile2: function nsDropbox_urlForFile(aFile, aCallback) {
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;
    this.requestObserver = aCallback;
//    aCallback.onStartRequest(null, null);
    if (!this._loggedIn) {
      var foo = this;
      return this.logon(function() {
        foo._getShareUrl(aFile, aCallback);
      });
    } else {
      this._getShareUrl(aFile, aCallback);
    }
  },
  
  _dropboxPath: null,

  _getDropboxDir: function() {
    if (this._dropboxPath) {
      var localFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
      localFile.initWithPath(this._dropboxPath);
      return localFile;
    }

    try
    {
      // see https://www.dropbox.com/developers/desktop_apps for details
      if (Services.appinfo.OS == "Darwin") {
        // Mac/Linux: ~/.dropbox/host.db
        var dropboxHostDB = Components.classes["@mozilla.org/file/directory_service;1"]
                                         .getService(Components.interfaces.nsIDirectoryServiceProvider)
                                         .getFile("Home",{});
        dropboxHostDB.append(".dropbox");
      } else {
        // Windows: %APPDATA%\Dropbox\host.db
        var dropboxHostDB = Components.classes["@mozilla.org/file/directory_service;1"]
                                         .getService(Components.interfaces.nsIDirectoryServiceProvider)
                                         .getFile("AppData",{});
        dropboxHostDB.append("Dropbox");
      }
      dropboxHostDB.append("host.db");
  
      var istream = Components.classes["@mozilla.org/network/file-input-stream;1"].
                    createInstance(Components.interfaces.nsIFileInputStream);
      istream.init(dropboxHostDB, 0x01, 0444, 0);
      istream.QueryInterface(Components.interfaces.nsILineInputStream);
  
      // the second line contains the path
      var line = {};
      istream.readLine(line);
      istream.readLine(line);
      istream.close();
  
      this._dropboxPath = atob(line.value);
    }
    catch (ex)
    {
      var dropboxDir = Components.classes["@mozilla.org/file/directory_service;1"]
                                 .getService(Components.interfaces.nsIDirectoryServiceProvider)
                                 .getFile("Home",{});
      dropboxDir.append("Dropbox");
      this._dropboxPath = dropboxDir.path;
    }
  
    var localFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
    localFile.initWithPath(this._dropboxPath);
    return localFile;
  },
  /**
   * For some nsILocalFile, check if it is a local Dropbox file
   *
   * @param aFile the nsILocalFile to check
   */
  isCloudFile: function nsDropbox_isCloudFile(aFile) {
    var dropboxDir = this._getDropboxDir();
    if (aFile.path.indexOf(dropboxDir.path) == 0) {
      return true;
    }
    return false;
  },

  /**
   * For some url, check if it is a Dropbox link
   *
   * @param aURL the url to check
   */
  isCloudURL: function nsDropbox_isCloudURL(aURL) {
    return aURL.toLowerCase().indexOf("http://db.tt") == 0;
  },

  /**
   * Updates the profile information for the account associated with the
   * account key.
   *
   * @param aWithUI a boolean for whether or not we should display authorization
   *                UI if we don't have a valid token anymore, or just fail out.
   * @param aCallback an nsIRequestObserver for observing the starting and
   *                  ending states of the request.
   */
  refreshUserInfo: function nsDropbox_refreshUserInfo(aWithUI, aCallback) {
    this.log.info("Getting User Info 1 : " + this._loggedIn);
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;
    this.requestObserver = aCallback;
    aCallback.onStartRequest(null, null);
    if (!this._loggedIn)
      return this._logonAndGetUserInfo(null, null, aWithUI);
    if (!this._userInfo)
      return this._getUserInfo();
    return this._userInfo;
  },


  /**
   * Our Dropbox implementation does not implement the createNewAccount
   * function defined in nsIMsgCloudFileProvider.idl.
   */
  createNewAccount: function nsDropbox_createNewAccount(aEmailAddress,
                                                        aPassword, aFirstName,
                                                        aLastName) {
    return Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  /**
   * If the user already has an account, we can get the user to just login
   * to it via OAuth.
   *
   * This function does not appear to be called from the BigFiles UI, and
   * might be excisable.
   */
  createExistingAccount: function nsDropbox_createExistingAccount(aRequestObserver) {
     // XXX: replace this with a better function
    let successCb = function(aResponseText, aRequest) {
      aRequestObserver.onStopRequest(null, this, Cr.NS_OK);
    }.bind(this);

    let failureCb = function(aResponseText, aRequest) {
      aRequestObserver.onStopRequest(null, this,
                                     Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    this.logon(successCb, failureCb, true);
  },

  /**
   * If the provider doesn't have an API for creating an account, perhaps
   * there's a url we can load in a content tab that will allow the user
   * to create an account.
   */
  get createNewAccountUrl() "",

  /**
   * For a particular error, return a URL if Dropbox has a page for handling
   * that particular error.
   *
   * @param aError the error to get the URL for
   */
  providerUrlForError: function nsDropbox_providerUrlForError(aError) {
    return "";
  },

  /**
   * If we don't know the limit, this will return -1.
   */
  get fileUploadSizeLimit() this._maxFileSize,
  get remainingFileSpace() this._totalStorage - this._fileSpaceUsed,
  get fileSpaceUsed() this._fileSpaceUsed,

  /**
   * Attempt to delete an upload file if we've uploaded it.
   *
   * @param aFile the file that was originall uploaded
   * @param aCallback an nsIRequestObserver for monitoring the starting and
   *                  ending states of the deletion request.
   */
  deleteFile: function nsDropbox_deleteFile(aFile, aCallback) {
    this.log.info("Deleting a file");

    if (Services.io.offline) {
      this.log.error("We're offline - we can't delete the file.");
      throw Ci.nsIMsgCloudFileProvider.offlineErr;
    }

    let uploadInfo = this._uploadInfo[aFile.path];
    if (!uploadInfo) {
      this.log.error("Could not find a record for the file to be deleted.");
      throw Cr.NS_ERROR_FAILURE;
    }

    let requestUrl = gServerUrl + kDeletePath + "&path=" + uploadInfo.path;;
    this.log.info("delete requestUrl = " + requestUrl);

    let deleteSuccess = function(aResponseText, aRequest) {
      // An empty 204 is sent on successful delete.
      this.log.info("delete request response = " + aRequest.status);

      if (aRequest.status != 204)
        aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
    }.bind(this);
    let deleteFailure = function(aException, aResponseText, aRequest) {
      this.log.error("Failed to delete file:" + aResponseText);
      aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
    }.bind(this);

    // Request to delete a file
    httpRequest(requestUrl, {
                  onLoad: deleteSuccess,
                  onError: deleteFailure,
                  method: "DELETE",
                  headers: [["Authorization", "Bearer " + this._oauth.accessToken]],
                });
  },

  /**
   * This function is used by our testing framework to override the default
   * URL's that nsDropbox connects to.
   */
  overrideUrls : function nsDropbox_overrideUrls(aNumUrls, aUrls) {
    gServerUrl = aUrls[0];
    gContentUrl = aUrls[1];
    gAuthUrl = aUrls[2];
    gLogoutUrl = aUrls[3];
  },

  /**
   * logon to the dropbox account.
   *
   * @param successCallback - called if logon is successful
   * @param failureCallback - called back on error.
   * @param aWithUI if false, logon fails if it would have needed to put up UI.
   *                This is used for things like displaying account settings,
   *                where we don't want to pop up the oauth ui.
   */
  logon: function nsDropbox_logon(successCallback, failureCallback, aWithUI) {
    // The token has expired, reauthenticate.
    if (this._oauth.tokenExpires < (new Date()).getTime()) {
      this._oauth.connect(successCallback, failureCallback, aWithUI);
     }
    // The token is still valid, success!
    else {
      successCallback();
    }
  },
  /**
   * Private function that attempts to retrieve the sharing URL for the file
   * uploaded with this Uploader.
   *
   * @param aFile ...
   * @param aCallback an nsIRequestObserver for monitoring the starting and
   *                  ending states of the URL retrieval request.
   */
  _getShareUrl: function nsDFU__getShareUrl(aFile, aCallback) {
    var self = this;
    var fileInfo = {
                     "settings": {
                       "requested_visibility": "public"
                      }
                   };
    var path = aFile.path.replace(this._dropboxPath, "");
    path = path.replace(/\\/gi,"/");
    fileInfo.path = path;
    let url = gServerUrl + kSharesPath;
    httpRequest(url, {
      onLoad: function(shareInfoJSON) {
        let shareInfo = JSON.parse(shareInfoJSON);
        let str = Cc["@mozilla.org/supports-string;1"]
                    .createInstance(Ci.nsISupportsString);
        str.data = shareInfo.url;
        aCallback.onStopRequest(null, str, Cr.NS_OK);
      },
      onError: function(a) {
        // Try getting the existing share link
        let url = gServerUrl + kListSharesPath;
        httpRequest(url, {
          onLoad: function(linksInfoJSON) {
            let linksInfo = JSON.parse(linksInfoJSON);
            if (linksInfo.length == 0) {
              self.log.error("Getting share URL failed: No links");
              aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
              return;
            }
            let shareInfo = linksInfo.links[0];
            let str = Cc["@mozilla.org/supports-string;1"]
                        .createInstance(Ci.nsISupportsString);
            str.data = shareInfo.url;
            aCallback.onStopRequest(null, str, Cr.NS_OK);
          },
          onError: function(a) {
            let errorInfo = JSON.parse(a);

            self.log.error("Getting share URL failed with response text: "
                           + a);
            aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
          },
          method: "POST",
          headers: [["Authorization", "Bearer " + self._oauth.accessToken], ["Content-Type", "application/json"]],
          postData: JSON.stringify({"path": path})
        });      },
      method: "POST",
      headers: [["Authorization", "Bearer " + this._oauth.accessToken], ["Content-Type", "application/json"]],
      postData: JSON.stringify(fileInfo)
    });
  },
};

function nsDropboxFileUploader(aDropbox, aFile, aCallback, aRequestObserver) {
  this.dropbox = aDropbox;
  this.log = this.dropbox.log;
  this.log.info("new nsDropboxFileUploader file = " + aFile.leafName);
  this.file = aFile;
  this.callback = aCallback;
  this.requestObserver = aRequestObserver;
}

nsDropboxFileUploader.prototype = {
  dropbox : null,
  file : null,
  callback : null,
  request : null,

  /**
   * Kicks off the upload request for the file associated with this Uploader.
   */
  uploadFile: function nsDFU_uploadFile() {
    this.requestObserver.onStartRequest(null, null);
    this.log.info("ready to upload file " + wwwFormUrlEncode(this.file.leafName));
    let url = gContentUrl + kFilesPutPath + 
              wwwFormUrlEncode(this.file.leafName) + "?overwrite=false";
    let fileContents = "";
    let fstream = Cc["@mozilla.org/network/file-input-stream;1"]
                     .createInstance(Ci.nsIFileInputStream);
    fstream.init(this.file, -1, 0, 0);
    let bufStream = Cc["@mozilla.org/network/buffered-input-stream;1"].
      createInstance(Ci.nsIBufferedInputStream);
    bufStream.init(fstream, this.file.fileSize);
    bufStream = bufStream.QueryInterface(Ci.nsIInputStream);
    let contentLength = fstream.available();
    let oauthParams =
      [];
    let uploadSuccess = function(aResponseText, aRequest) {
      this.request = null;
      this.log.info("success putting file " + aResponseText);
      let putInfo = JSON.parse(aResponseText);
      this.dropbox._uploadInfo[this.file.path] = putInfo;
      this.dropbox._getShareUrl(this.file, this.callback);
    }.bind(this);
    let uploadFailure = function(aException, aResponseText, aRequest) {
        this.request = null;
        this.log.info("failed putting file response = " + aResponseText);
        if (this.callback)
          this.callback(this.requestObserver,
                        Ci.nsIMsgCloudFileProvider.uploadErr);
    }.bind(this);

    // Request to delete a file
    this.request = httpRequest(url, {
                  onLoad: uploadSuccess,
                  onError: uploadFailure,
                  method: "POST",
                  headers: [["Authorization", "Bearer " + this.dropbox._oauth.accessToken],
                            ["Content-Length", contentLength]],
                });

  },

  /**
   * Cancels the upload request for the file associated with this Uploader.
   */
  cancel: function nsDFU_cancel() {
    this.callback(this.requestObserver, Ci.nsIMsgCloudFileProvider.uploadCanceled);
    if (this.request) {
      let req = this.request;
      if (req.channel) {
        this.log.info("canceling channel upload");
        delete this.callback;
        req.channel.cancel(Cr.NS_BINDING_ABORTED);
      }
      this.request = null;
    }
  },
};

const NSGetFactory = XPCOMUtils.generateNSGetFactory([nsDropbox]);
