/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

/* This file implements the nsIMsgCloudFileProvider interface.
 *
 * This component handles the OneDrive implementation of the
 * nsIMsgCloudFileProvider interface.
 */

const {classes: Cc, interfaces: Ci, utils: Cu, results: Cr} = Components;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource:///modules/gloda/log4moz.js");
Cu.import("resource:///modules/cloudFileAccounts.js");
Cu.import("resource:///modules/OAuth2.jsm");
Cu.import("resource://gre/modules/Http.jsm");

var gServerUrl = "https://apis.live.net/v5.0/";
var gUploadUrl = "https://apis.live.net/v5.0/";

const kClientId = "000000004C146709";
const kClientSecret = "VGG9a-om0G6HosDjm266Rw1UqOjnZXae";
const kAuthBaseUrl = "https://login.live.com/oauth20_";
const kScope = "wl.skydrive_update wl.offline_access";
const kAuthUrl = "authorize.srf";
const kTokenUrl = "token.srf";

function nsOneDrive() {
  this.log = Log4Moz.getConfiguredLogger("OneDriveService");
  this.log.level = Log4Moz.Level.Debug;
  this._oauth = new OAuth2(kAuthBaseUrl, kScope, kClientId, kClientSecret);
  this._oauth.authURI = kAuthBaseUrl + kAuthUrl;
  this._oauth.tokenURI = kAuthBaseUrl + kTokenUrl;
  this._oauth.completionURI = "http://localhost.postbox";

  let account = this;
  Object.defineProperty(this._oauth, "refreshToken", {
    get: function getRefreshToken() {
      if (!this.mRefreshToken) {
        let authToken = cloudFileAccounts.getSecretValue(account.accountKey,
                                                         cloudFileAccounts.kTokenRealm);
        this.mRefreshToken = authToken || "";
      }
      return this.mRefreshToken;
    },
    set: function setRefreshToken(aVal) {
      if (!aVal)
        aVal = "";

      cloudFileAccounts.setSecretValue(account.accountKey,
                                       cloudFileAccounts.kTokenRealm,
                                       aVal);

      return (this.mRefreshToken = aVal);
    },
    enumerable: true
  });
}

nsOneDrive.prototype = {
  /* nsISupports */
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIMsgCloudFileProvider, Ci.nsIMsgCloudFileProvider2]),

  classID: Components.ID("{c529b020-950b-11e4-b4a9-0800200c9a66}"),

  get type() "OneDrive",
  get displayName() "OneDrive",
  get serviceURL() "https://onedrive.live.com",
  get iconClass() "chrome://pbfilelink/content/onedrive/onedrive.png",
  get accountKey() this._accountKey,
  get lastError() this._lastErrorText,
  get settingsURL() "chrome://pbfilelink/content/onedrive/settings.xhtml",
  get managementURL() "chrome://pbfilelink/content/onedrive/management.xhtml",

  completionURI: "http://localhost.postbox",

  _accountKey: false,
  _prefBranch: null,
  _folderId: "",
  // If an access token exists, the user is logged in.
  get _loggedIn() !!this._oauth.accessToken,
  _userInfo: null,
  _file : null,
  _maxFileSize : 2147483648, // Microsoft says he max is 2GB
  _fileSpaceUsed : -1,
  _totalStorage : -1,
  _lastErrorStatus : 0,
  _lastErrorText : "",
  _uploadingFile : null,
  _uploader : null,
  _urlsForFiles : {},
  _uploadInfo : {},
  _uploads: [],
  _oauth: null,

  /**
   * Used by our testing framework to override the URLs that this component
   * communicates to.
   */
  overrideUrls: function nsOneDrive_overrideUrls(aNumUrls, aUrls) {
    gServerUrl = aUrls[0];
  },

  /**
   * Initializes an instance of this nsIMsgCloudFileProvider for an account
   * with key aAccountKey.
   *
   * @param aAccountKey the account key to initialize this
   *                    nsIMsgCloudFileProvider with.
   */
  init: function nsOneDrive_init(aAccountKey) {
    this._accountKey = aAccountKey;
    this._prefBranch = Services.prefs.getBranch("mail.cloud_files.accounts." +
                                                aAccountKey + ".");
  },

  /**
   * Private function for assigning the folder id from a cached version
   * If the folder doesn't exist, set in motion the creation
   *
   * @param aCallback called if folder is ready.
   */
  _initFolder: function nsOneDrive__initFolder(aCallback) {
    this.log.info('_initFolder, cached folder id  = ' + this._cachedFolderId);

    let saveFolderId = function(aFolderId) {
      this.log.info('saveFolderId : ' + aFolderId);
      this._cachedFolderId = this._folderId = aFolderId;
      if (aCallback)
        aCallback();
    }.bind(this);

    let createThunderbirdFolder = function() {
      this._createFolder("Thunderbird", saveFolderId);
    }.bind(this);

    if (this._cachedFolderId == "")
      createThunderbirdFolder();
    else {
      this._folderId = this._cachedFolderId;
      if (aCallback)
        aCallback();
    }
  },

  /**
   * Private callback function passed to, and called from
   * nsOneDriveFileUploader.
   *
   * @param aRequestObserver a request observer for monitoring the start and
   *                         stop states of a request.
   * @param aStatus the status of the request.
   */
  _uploaderCallback: function nsOneDrive__uploaderCallback(aRequestObserver,
                                                            aStatus) {
    aRequestObserver.onStopRequest(null, null, aStatus);
    this._uploadingFile = null;
    this._uploads.shift();
    if (this._uploads.length > 0) {
      let nextUpload = this._uploads[0];
      this.log.info("chaining upload, file = " + nextUpload.file.leafName);
      this._uploadingFile = nextUpload.file;
      this._uploader = nextUpload;
      try {
        this.uploadFile(nextUpload.file, nextUpload.requestObserver);
      }
      catch (ex) {
        nextUpload.callback(nextUpload.requestObserver, Cr.NS_ERROR_FAILURE);
      }
    }
    else
      this._uploader = null;
  },

  /**
   * Attempt to upload a file to OneDrive's servers.
   *
   * @param aFile an nsILocalFile for uploading.
   * @param aCallback an nsIRequestObserver for monitoring the start and
   *                  stop states of the upload procedure.
   */
  uploadFile: function nsOneDrive_uploadFile(aFile, aCallback) {
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;

    this.log.info("uploading " + aFile.leafName);

    // Some ugliness here - we stash requestObserver here, because we might
    // use it again in _getUserInfo.
    this.requestObserver = aCallback;

    // if we're uploading a file, queue this request.
    if (this._uploadingFile && this._uploadingFile != aFile) {
      let uploader = new nsOneDriveFileUploader(this, aFile,
                                               this._uploaderCallback
                                                   .bind(this),
                                               aCallback);
      this._uploads.push(uploader);
      return;
    }
    this._file = aFile;
    this._uploadingFile = aFile;

    let finish = function() {
      this._finishUpload(aFile, aCallback);
    }.bind(this);

    let onGetUserInfoSuccess = function() {
      this._initFolder(finish);
    }.bind(this);

    let onAuthFailure = function() {
      this._urlListener.onStopRequest(null, null,
                                      Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    this.log.info("Checking to see if we're logged in");

    if (!this._loggedIn) {
      let onLoginSuccess = function() {
        this._getUserInfo(onGetUserInfoSuccess, onAuthFailure);
      }.bind(this);

      return this.logon(onLoginSuccess, onAuthFailure, true);
    }

    if (!this._userInfo)
      return this._getUserInfo(onGetUserInfoSuccess, onAuthFailure);

    onGetUserInfoSuccess();
  },

  /**
   * A private function called when we're almost ready to kick off the upload
   * for a file. First, ensures that the file size is not too large, and that
   * we won't exceed our storage quota, and then kicks off the upload.
   *
   * @param aFile the nsILocalFile to upload
   * @param aCallback the nsIRequestObserver for monitoring the start and stop
   *                  states of the upload procedure.
   */
  _finishUpload: function nsOneDrive__finishUpload(aFile, aCallback) {
    let exceedsFileLimit = Ci.nsIMsgCloudFileProvider.uploadExceedsFileLimit;
    let exceedsQuota = Ci.nsIMsgCloudFileProvider.uploadWouldExceedQuota;
    if (aFile.fileSize > this._maxFileSize)
      return aCallback.onStopRequest(null, null, exceedsFileLimit);
    if (aFile.fileSize > this.remainingFileSpace)
      return aCallback.onStopRequest(null, null, exceedsQuota);

    delete this._userInfo; // force us to update userInfo on every upload.

    if (!this._uploader) {
      this._uploader = new nsOneDriveFileUploader(this, aFile,
                                             this._uploaderCallback
                                                 .bind(this),
                                             aCallback);
      this._uploads.unshift(this._uploader);
    }

    this._uploadingFile = aFile;
    this._uploader.uploadFile();
  },

  /**
   * Cancels an in-progress file upload.
   *
   * @param aFile the nsILocalFile being uploaded.
   */
  cancelFileUpload: function nsOneDrive_cancelFileUpload(aFile) {
    if (this._uploadingFile.equals(aFile)) {
      this._uploader.cancel();
    }
    else {
      for (let i = 0; i < this._uploads.length; i++)
        if (this._uploads[i].file.equals(aFile)) {
          this._uploads[i].requestObserver.onStopRequest(
            null, null, Ci.nsIMsgCloudFileProvider.uploadCanceled);
          this._uploads.splice(i, 1);
          return;
        }
    }
  },

  /**
   * A private function for retrieving profile information about a user.
   *
   * @param successCallback a callback fired if retrieving profile information
   *                        is successful.
   * @param failureCallback a callback fired if retrieving profile information
   *                        fails.
   */
  _getUserInfo: function nsOneDrive__getUserInfo(successCallback, failureCallback) {
    let requestUrl = gServerUrl + "me/skydrive/quota";
    this.log.info("get_account_info requestUrl = " + requestUrl);

    if (!successCallback)
      successCallback = function() {
        this.requestObserver
            .onStopRequest(null, null,
                           this._loggedIn ? Cr.NS_OK : Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    if (!failureCallback)
      failureCallback = function () {
        this.requestObserver
            .onStopRequest(null, null, Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    let accountInfoSuccess = function(aResponseText, aRequest) {
      this.log.info("get_account_info request response = " + aResponseText);

      try {
        this._userInfo = JSON.parse(aResponseText);

        if (!this._userInfo) {
          this.failureCallback();
          return;
        }

        this._totalStorage = this._userInfo.quota;
        this._fileSpaceUsed = this._userInfo.quota - this._userInfo.available;
        this.log.info("storage total = " + this._totalStorage);
        this.log.info("storage used = " + this._fileSpaceUsed);
        successCallback();
      }
      catch(e) {
        // most likely bad JSON
        this.log.error("Failed to parse account info response: " + e);
        this.log.error("Account info response: " + aResponseText);
        failureCallback();
      }
    }.bind(this);
    let accountInfoFailure = function(aException, aResponseText, aRequest) {
      this.log.info("Failed to acquire user info:" + aResponseText);
      this.log.error("user info failed, status = " + aRequest.status);
      this.log.error("response text = " + aResponseText);
      this.log.error("exception = " + aException);
      failureCallback();
    }.bind(this)

    // Request to get user info
    httpRequest(requestUrl, {
                  onLoad: accountInfoSuccess,
                  onError: accountInfoFailure,
                  method: "GET",
                  headers: [["Authorization", "Bearer " + this._oauth.accessToken]]
                });
  },

  /**
   * A private function that first ensures that the user is logged in, and then
   * retrieves the user's profile information.
   *
   * @param aSuccessCallback the function called on successful information
   *                         retrieval
   * @param aFailureCallback the function called on failed information retrieval
   * @param aWithUI a boolean for whether or not we should display authorization
   *                UI if we don't have a valid token anymore, or just fail out.
   */
  _logonAndGetUserInfo: function nsOneDrive_logonAndGetUserInfo(aSuccessCallback,
                                                               aFailureCallback,
                                                               aWithUI) {
    if (!aFailureCallback)
      aFailureCallback = function () {
        this.requestObserver
            .onStopRequest(null, null, Ci.nsIMsgCloudFileProvider.authErr);
      }.bind(this);

    return this.logon(function() {
      this._getUserInfo(aSuccessCallback, aFailureCallback);
    }.bind(this), aFailureCallback, aWithUI);
  },

  /**
   * Returns the sharing URL for some uploaded file.
   *
   * @param aFile the nsILocalFile to get the URL for.
   */
  urlForFile: function nsOneDrive_urlForFile(aFile) {
    return this._urlsForFiles[aFile.path];
  },

  /**
   * For some nsILocalFile, return the associated sharing URL.
   *
   * @param aFile the nsILocalFile to retrieve the URL for
   */
  urlForFile2: function nsOneDrive_urlForFile(aFile, aCallback) {
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;
    this.requestObserver = aCallback;
//    aCallback.onStartRequest(null, null);
    if (!this._loggedIn) {
      var foo = this;
      return this.logon(function() {
        foo._getShareUrl(aFile, aCallback);
      });
    } else {
      this._getShareUrl(aFile, aCallback);
    }
  },

  /**
   * Private function that attempts to retrieve the sharing URL for the file
   * uploaded with this Uploader.
   *
   * @param aFile ...
   * @param aCallback an nsIRequestObserver for monitoring the starting and
   *                  ending states of the URL retrieval request.
   */
  _getShareUrl: function nsOneDrive_getShareUrl(aFile, aCallback) {
    var oa = this._oauth;
    onDriveGetFileID(oa, aFile.path, function(id) {
      httpRequest("https://apis.live.net/v5.0/" + id + "/shared_read_link", {
        onLoad: function(linkInfoJSON) {
          var linkInfo = JSON.parse(linkInfoJSON);
          let str = Cc["@mozilla.org/supports-string;1"]
                      .createInstance(Ci.nsISupportsString);
          str.data = linkInfo.link;
          aCallback.onStopRequest(null, str, Cr.NS_OK);
        },
        onError: function(a) {
          this.log.error("Getting share URL failed with response text: "
                         + aResponseText);
          aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
        },
        method: "GET",
        headers: [["Authorization", "Bearer " + oa.accessToken]]
        });
    });
  },

  /**
   * For some nsILocalFile, check if it is a local Dropbox file
   *
   * @param aFile the nsILocalFile to check
   */
  isCloudFile: function nsOneDrive_isCloudFile(aFile) {
    var parent = aFile;
    while ((parent = parent.parent)) {
      if (parent.leafName == "OneDrive") {
        return true;
      }
    }
    return false;
  },

  /**
   * For some url, check if it is a Dropbox link
   *
   * @param aURL the url to check
   */
  isCloudURL: function nsOneDrive_isCloudURL(aURL) {
    return aURL.toLowerCase().indexOf("https://onedrive.live.com/") == 0;
  },

  /**
   * Updates the profile information for the account associated with the
   * account key.
   *
   * @param aWithUI a boolean for whether or not we should display authorization
   *                UI if we don't have a valid token anymore, or just fail out.
   * @param aCallback an nsIRequestObserver for observing the starting and
   *                  ending states of the request.
   */
  refreshUserInfo: function nsOneDrive_refreshUserInfo(aWithUI, aCallback) {
    this.log.info("Getting User Info 1 : " + this._loggedIn);
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;
    this.requestObserver = aCallback;
    aCallback.onStartRequest(null, null);
    if (!this._loggedIn)
      return this._logonAndGetUserInfo(null, null, aWithUI);
    if (!this._userInfo)
      return this._getUserInfo();
    return this._userInfo;
  },

  /**
   * Our OneDrive implementation does not implement the createNewAccount
   * function defined in nsIMsgCloudFileProvider.idl.
   */
  createNewAccount: function nsOneDrive_createNewAccount(aEmailAddress,
                                                          aPassword, aFirstName,
                                                          aLastName) {
    return Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  /**
   * Private function for creating folder on the OneDrive website.
   *
   * @param aName name of folder
   * @param aSuccessCallback called when folder is created
   */
  _createFolder: function nsOneDrive__createFolder(aName,
                                              aSuccessCallback) {
    this.log.info("Creating folder: " + aName);
    if (Services.io.offline)
      throw Ci.nsIMsgCloudFileProvider.offlineErr;

    let body = {
      name: aName
    };
    let requestUrl = gServerUrl + "me/skydrive";
    this.log.info("create_folder requestUrl = " + requestUrl);

    let createSuccess = function(aResponseText, aRequest) {
      this.log.info("create_folder request response = " + aResponseText);

      try {
        let result = JSON.parse(aResponseText);

        if (!result || !result.id) {
          this._lastErrorText = "Create folder failure";
          this._lastErrorStatus = docStatus;
          return;
        }
        let folderId = result.id;
        this.log.info("folder id = " + folderId);
        aSuccessCallback(folderId);
      }
      catch(e) {
        // most likely bad JSON
        this.log.error("Failed to create a new folder");
      }
    }.bind(this);
    let createFailure = function(aException, aResponseText, aRequest) {
      this.log.error("Failed to create a new folder: " + aRequest.status);
    }.bind(this);

    // Request to create the folder
    httpRequest(requestUrl, {
                  onLoad: createSuccess,
                  onError: createFailure,
                  method: "POST",
                  headers: [["Authorization", "Bearer " + this._oauth.accessToken],
                            ["Content-Type", "application/json"]],
                  postData: JSON.stringify(body)
                });
  },

  /**
   * If a the user associated with this account key already has an account,
   * allows them to log in.
   *
   * @param aRequestObserver an nsIRequestObserver for monitoring the start and
   *                         stop states of the login procedure.
   */
  createExistingAccount: function nsOneDrive_createExistingAccount(aRequestObserver) {
     // XXX: replace this with a better function
    let successCb = function(aResponseText, aRequest) {
      aRequestObserver.onStopRequest(null, this, Cr.NS_OK);
    }.bind(this);

    let failureCb = function(aResponseText, aRequest) {
      aRequestObserver.onStopRequest(null, this,
                                     Ci.nsIMsgCloudFileProvider.authErr);
    }.bind(this);

    this.logon(successCb, failureCb, true);
  },

  /**
   * Returns an appropriate provider-specific URL for dealing with a particular
   * error type.
   *
   * @param aError an error to get the URL for.
   */
  providerUrlForError: function nsOneDrive_providerUrlForError(aError) {
    if (aError == Ci.nsIMsgCloudFileProvider.uploadWouldExceedQuota)
      return "https://onedrive.live.com/about/en-us/plans/";
    return "";
  },

  /**
   * If the provider doesn't have an API for creating an account, perhaps
   * there's a url we can load in a content tab that will allow the user
   * to create an account.
   */
  get createNewAccountUrl() "",

  /**
   * If we don't know the limit, this will return -1.
   */
  get fileUploadSizeLimit() this._maxFileSize,
  get remainingFileSpace() this._totalStorage - this._fileSpaceUsed,
  get fileSpaceUsed() this._fileSpaceUsed,

  /**
   * Attempts to delete an uploaded file.
   *
   * @param aFile the nsILocalFile to delete.
   * @param aCallback an nsIRequestObserver for monitoring the start and stop
   *                  states of the delete procedure.
   */
  deleteFile: function nsOneDrive_deleteFile(aFile, aCallback) {
    this.log.info("Deleting a file");

    if (Services.io.offline) {
      this.log.error("We're offline - we can't delete the file.");
      throw Ci.nsIMsgCloudFileProvider.offlineErr;
    }

    let uploadInfo = this._uploadInfo[aFile.path];
    if (!uploadInfo) {
      this.log.error("Could not find a record for the file to be deleted.");
      throw Cr.NS_ERROR_FAILURE;
    }

    let requestUrl = gServerUrl + uploadInfo.fileId;
    this.log.info("delete requestUrl = " + requestUrl);

    let deleteSuccess = function(aResponseText, aRequest) {
      // An empty 204 is sent on successful delete.
      this.log.info("delete request response = " + aRequest.status);

      if (aRequest.status != 204)
        aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
    }.bind(this);
    let deleteFailure = function(aException, aResponseText, aRequest) {
      this.log.error("Failed to delete file:" + aResponseText);
      aCallback.onStopRequest(null, null, Cr.NS_ERROR_FAILURE);
    }.bind(this);

    // Request to delete a file
    httpRequest(requestUrl, {
                  onLoad: deleteSuccess,
                  onError: deleteFailure,
                  method: "DELETE",
                  headers: [["Authorization", "Bearer " + this._oauth.accessToken]],
                });
  },

  /**
   * Attempt to log on and get the auth token for this OneDrive account.
   *
   * @param successCallback the callback to be fired if logging on is successful
   * @param failureCallback the callback to be fired if loggong on fails
   * @aparam aWithUI a boolean for whether or not we should prompt for a password
   *                 if no auth token is currently stored.
   */
  logon: function nsOneDrive_logon(successCallback, failureCallback, aWithUI) {
    // The token has expired, reauthenticate.
    if (this._oauth.tokenExpires < (new Date()).getTime()) {
      this._oauth.connect(successCallback, failureCallback, aWithUI);
     }
    // The token is still valid, success!
    else {
      successCallback();
    }
  },

  get _cachedFolderId() {
    let folderId = "";
    try {
      folderId = this._prefBranch.getCharPref("folderid");
    }
    catch(e) { } // pref does not exist

    return folderId;
  },

  set _cachedFolderId(aVal) {
    if (!aVal)
      aVal = "";

    this._prefBranch.setCharPref("folderid", aVal);
  },
};

function nsOneDriveFileUploader(aOneDrive, aFile, aCallback,
                                 aRequestObserver) {
  this.onedrive = aOneDrive;
  this.log = this.onedrive.log;
  this.log.info("new nsOneDriveFileUploader file = " + aFile.leafName);
  this.file = aFile;
  this.callback = aCallback;
  this.requestObserver = aRequestObserver;
}

nsOneDriveFileUploader.prototype = {
  onedrive : null,
  file : null,
  callback : null,
  request : null,

  /**
   * Do the upload of the file to OneDrive.
   */
  uploadFile: function nsOneDrive_uploadFile() {
    this.requestObserver.onStartRequest(null, null);
    let requestUrl = gUploadUrl + this.onedrive._cachedFolderId + "/files";
    this.onedrive._uploadInfo[this.file.path] = {};
    this.onedrive._uploadInfo[this.file.path].uploadUrl = requestUrl;

    let req = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"]
                .createInstance(Ci.nsIXMLHttpRequest);

    let curDate = Date.now().toString();
    this.log.info("upload url = " + requestUrl);
    this.request = req;
    req.open("POST", requestUrl, true);
    req.onload = function() {
      if (req.status >= 200 && req.status < 400) {
        try {
          this.log.info("upload response = " + req.responseText);
          let result = JSON.parse(req.responseText);

          if (result.id) {
            // Request a shared link for this file.
            let shareSuccess = function(aResponseText, aRequest) {
              this.log.info("share response = " + aResponseText);

              let result = JSON.parse(aResponseText);
              // If shared_link doesn't exist or is empty, an error occurred.
              if (!result.link) {
                this.callback(this.requestObserver,
                  Ci.nsIMsgCloudFileProvider.uploadErr);
              }

              // Can't use download_url because free accounts do not have access
              // to provide direct download URLs.
              let url = result.link;
              this.log.info("public_name = " + url);
              this.onedrive._urlsForFiles[this.file.path] = url;

              this.callback(this.requestObserver, Cr.NS_OK);
            }.bind(this);
            let shareFailure = function(aResponseText, aReqest) {
              this.callback(this.requestObserver,
                Ci.nsIMsgCloudFileProvider.uploadErr);
            }.bind(this);

            // Currently only one file is uploaded at a time.
            let fileId = result.id;
            this.onedrive._uploadInfo[this.file.path].fileId = fileId;

            let requestUrl = gServerUrl + fileId + "/shared_read_link";
            httpRequest(requestUrl, {
              onLoad: shareSuccess,
              onError: shareFailure,
              method: "GET",
              headers: [["Authorization", "Bearer " + this.onedrive._oauth.accessToken]]
            });
          }
          else {
            this.callback(this.requestObserver,
                      Ci.nsIMsgCloudFileProvider.uploadErr);
          }
        } catch (ex) {
          this.log.error(ex);
          this.callback(this.requestObserver,
                      Ci.nsIMsgCloudFileProvider.uploadErr);
        }
      }
      else {
        this.callback(this.requestObserver,
                      Ci.nsIMsgCloudFileProvider.uploadErr);
      }
    }.bind(this);

    req.onerror = function (a) {
      if (this.callback)
        this.callback(this.requestObserver,
                      Ci.nsIMsgCloudFileProvider.uploadErr);
    }.bind(this);

    req.setRequestHeader("Authorization", "Bearer " + this.onedrive._oauth.accessToken);

    // Encode the form.
    let file = new File(this.file);
    let form = Cc["@mozilla.org/files/formdata;1"]
                 .createInstance(Ci.nsIDOMFormData);
    form.append("file", file, this.file.leafName);

    req.send(form);
  },

  /**
   * Cancels the upload request for the file associated with this Uploader.
   */
  cancel: function nsOneDrive_cancel() {
    this.log.info("in uploader cancel");
    this.callback(this.requestObserver, Ci.nsIMsgCloudFileProvider.uploadCanceled);
    delete this.callback;
    if (this.request) {
      this.log.info("cancelling upload request");
      let req = this.request;
      if (req.channel) {
        this.log.info("cancelling upload channel");
        req.channel.cancel(Cr.NS_BINDING_ABORTED);
      }
      this.request = null;
    }
  }
};

const NSGetFactory = XPCOMUtils.generateNSGetFactory([nsOneDrive]);

function onDriveGetFileID(oa, path, aCallback) {
  var ucn = Components.classes["@mozilla.org/intl/unicodenormalizer;1"].getService(Components.interfaces.nsIUnicodeNormalizer);
  var cvtpath = {};
  ucn.NormalizeUnicodeNFKC(path, cvtpath);

  // Remove everything before "OneDrive";
  var breakup = cvtpath.value.split("OneDrive");
  var newpath = "OneDrive" + breakup[1];
  newpath = newpath.replace(/\\/gi,"/");
  var paths = newpath.split("/");
  
  
  httpRequest("https://apis.live.net/v5.0/me/skydrive/search?q=" + encodeURIComponent(paths[paths.length-1]), {
    onLoad: function(fileinfoJSON) {
      var files = JSON.parse(fileinfoJSON).data;
      if (files.length == 0) {
        // File doesn't exist. Shouldn't happen.
        aCallback("");
        return;
      }
      if (files.length > 1) {
        // Remove any items from the list that don't match the exact filename.
        files = files.filter(function(file) {
          return file.name == paths[paths.length-1];
        });
      }
      if (files.length == 1) {
        // Only one copy. Return it.
        aCallback(files[0].id);
        return;
      }
      // Check the parent folders and see if they match the name of the parent
      // If they do, keep going up the chain until nothing.
      files.forEach(function(file) {
        checkParentPath(oa, file.parent_id, file.id, 1, paths, aCallback);
      });
    },
    onError: function(e) {
      Components.utils.reportError(e);
    },
    method: "GET",
    headers: [["Authorization", "Bearer " + oa.accessToken]]
  });
}

function checkParentPath(oa, parentID, origFileID, nestLevel, origPath, aCallback) {
  httpRequest("https://apis.live.net/v5.0/" + parentID, {
  onLoad: function(folderInfoJSON) {
      var folder = JSON.parse(folderInfoJSON);
      if (!folder.parent_id && origPath.length-1 == nestLevel) {
        aCallback(origFileID);
        return;
      }
      if (origPath[origPath.length-(nestLevel+1)] != folder.name &&
          (folder.name != "SkyDrive" && origPath[origPath.length-(nestLevel+1)] != "OneDrive")) { // Stupid workaround - internal name is still SkyDrive
        return;
      }
      if (folder.parent_id) {
        checkParentPath(oa, folder.parent_id, origFileID, nestLevel+1, origPath, aCallback);
      }
    },
    onError: function(e) {
      // Top level directory
      // Still might be sucesss if that's what they asked for
      if (origPath.length == nestLevel) {
        aCallback(origFileID);
        return;
      }
    },
    method: "GET",
    headers: [["Authorization", "Bearer " + oa.accessToken]]
  });
}
